# @DEPRECATED

# This file requires the following dependencies
# "NEURON>=8.2.0",
# "numpy==1.23.5",
# "sbi==0.17.2",
# "matplotlib",
# "torch",
# "scikit-learn"

import json

import numpy as np
import torch
from neuron import h
from sbi import utils
from sbi.inference import SNPE, prepare_for_sbi, simulate_for_sbi
from scipy.signal import resample
from sklearn.metrics import mean_squared_error
from sklearn.tree import DecisionTreeRegressor

from act.cell import Cell


class ACTOptimizer:
    """
    Base class ACT-compatible optimizers should conform to.
    """

    def __init__(self, config_file: str) -> None:
        """
        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """

        # "general"
        self.cell = None

        # "run_options"
        self.v_init = None
        self.tstop = None
        self.i_clamp_delay = None
        self.i_clamp_dur = None

        # "optimization_parameters"
        self.current_injections = None
        self.parameters = None
        self.lows = None
        self.highs = None

        # Load the standard hoc file
        h.load_file("stdrun.hoc")

        # Parse config
        self.parse_config(config_file=config_file)

        # For convenience
        self.num_current_injections = len(self.current_injections)
        self.num_parameters = len(self.parameters)

        # Set current clamp session
        self.i_clamp = h.IClamp(self.cell.get_recording_section())

        # Observed (target) data
        self.observed_data = None

    def optimize(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        num_epochs: int,
        num_simulations: int,
    ) -> np.array:
        """
        Optimize parameters. This is the optimizer's main method which should
        (1) train the feature model if applicable,
        (2) perform the optimization,
        (3) return parameters's estimates.

        Additional arguments can be used in specific implementations.

        Parameters:
        ----------
        feature_model: torch.nn.Module
            Model that generates summary features.

        num_epochs: int
            Number of epochs (rounds) to run.

        num_simulations: int
            Number of simulations to draw each round.

        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        Returns:
        ----------
        estimates: np.array
            Parameter estimates made for observed_data.

        """
        raise NotImplementedError

    def optimize_with_segregation(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        parameter_inds: list,
        voltage_bounds: list,
        num_epochs: int,
        num_simulations: int,
    ) -> np.array:
        """
        Optimize using the segregation scheme from Alturki2016. Segregates and runs self.optimize(...).

        Parameters:
        ----------
        feature_model: torch.nn.Module
            Model that generates summary features.

        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        parameter_inds: list
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        num_epochs: int
            Number of epochs (rounds) to run.

        num_simulations: int
            Number of simulations to draw each round.

        Returns:
        ----------
        estimates: np.array
            Parameter estimates made for observed_data.

        """
        raise NotImplementedError

    def prepare_for_segregation(
        self, observed_data: torch.Tensor, parameter_inds: list, voltage_bounds: list
    ):
        """
        Cuts a specified region from the voltage trace and resamples the cut portions to match the original data's length
        and sets parameter lists to optimize for user-defined parameters.

        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        parameter_inds: list[int]
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        Returns:
        ----------
        cut_observed_data: torch.Tensor
            Data to estimate on.

        original_param_set: list[list]
            List of original values for self.parameters, self.lows, self.highs, self.num_parameters to use for a reset later.
        """

        # Cut observed data and temporarily drop parameters which are not in the region of interest
        # Clunky, but it is a bad side-effect of using config files.
        cut_observed_data = torch.zeros_like(observed_data)
        for i in range(self.num_current_injections):
            trace = observed_data[i]
            trace = trace[(trace >= voltage_bounds[0]) & (trace < voltage_bounds[1])]
            cut_observed_data[i] = torch.tensor(
                resample(x=trace, num=observed_data.shape[1])
            ).float()

        original_param_set = [
            self.parameters.copy(),
            self.lows.copy(),
            self.highs.copy(),
            self.num_parameters,
        ]
        self.parameters = [self.parameters[i] for i in parameter_inds]
        self.lows = [self.lows[i] for i in parameter_inds]
        self.highs = [self.highs[i] for i in parameter_inds]
        self.num_parameters = len(self.parameters)

        return cut_observed_data, original_param_set

    def simulate(self, parameters: np.ndarray) -> torch.Tensor:
        """
        Function to simulate data from a model for optimizer to use.

        Parameters:
        ----------
        parameters: ndarray(shape = num_parameters)
            Parameters optimized for.

        Returns:
        ----------
        out: ndarray(shape = (num_current_injections, tstop * 10 + 1))
            Generated data.
        """
        simulated_data = []

        for inj in self.current_injections:
            # Set simulation parameters
            steps_per_ms = 10
            h.steps_per_ms = steps_per_ms
            h.dt = 1 / steps_per_ms
            h.tstop = self.tstop
            h.v_init = self.v_init

            self.i_clamp.dur = self.i_clamp_dur
            self.i_clamp.amp = inj
            self.i_clamp.delay = self.i_clamp_delay

            self.cell.set_parameters(self.parameters, parameters)

            # Run the simulation with the given parameters
            h.run()

            voltage = self.cell.mem_potential.as_numpy()
            simulated_data.append(voltage)

        out = np.array(simulated_data)

        return out

    def parse_config(self, config_file: str) -> None:
        """
        Parse the configuration files and fill the attributes.

        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """

        with open(config_file) as file:
            config_data = json.load(file)

        self.cell = Cell(
            hoc_file=config_data["general"]["hoc_file"],
            cell_name=config_data["general"]["cell_name"],
        )

        run_options = config_data["run_options"]
        self.v_init = run_options["v_init"]
        self.tstop = run_options["tstop"]
        self.i_clamp_delay = run_options["i_clamp_delay"]
        self.i_clamp_dur = run_options["i_clamp_dur"]

        optimization_parameters = config_data["optimization_parameters"]
        self.current_injections = optimization_parameters["current_injections"]
        self.parameters = optimization_parameters["parameters"]
        self.lows = optimization_parameters["lows"]
        self.highs = optimization_parameters["highs"]


class SBIOptimizer(ACTOptimizer):
    """
    Optimizer based on SBI's SNPE-C (https://www.mackelab.org/sbi/). Trains the feature model at each epoch.
    """

    def __init__(self, config_file: str) -> None:
        """
        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """

        super().__init__(config_file=config_file)

        # Set prior
        self.lows = torch.tensor(self.lows, dtype=float)
        self.highs = torch.tensor(self.highs, dtype=float)
        self.prior = utils.BoxUniform(low=self.lows, high=self.highs)

        # Set posterior
        self.posterior = []

    def optimize(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        num_epochs: int = 1,
        num_simulations: int = 100,
        num_samples: int = 10,
        workers: int = 1,
        verbose=False,
    ) -> np.array:
        """
        Parameters:
        ----------
        feature_model: torch.nn.Module
            Model that generates summary features. Accepts tensors of shape (num_current_injections, len_voltage_trace).
            Returns tensors of shape (num_current_injections, num_summary_features).

        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        num_epochs: int = 1
            Number of epochs (rounds) to run.

        num_simulations: int = 100
            Number of simulations to draw each round.

        num_samples: int = 10
            Number of parameter estimates' samples to return.

        workers: int = 1
            Number of threads for SBI to use.

        verbose: bool = False
            Show progress information reported by SBI.

        Returns:
        ----------
        estimates: np.array, shape = (num_current_injections, num_samples, number of parameters)
            Parameter estimates made for observed_data.
        """

        self.observed_data = observed_data

        # Prepare the simulation function and prior for SBI (default method from SBI)
        # Simulation function must return tensors of shape (num_current_injections, voltage_trace_length)
        self.simulator, self.prior = prepare_for_sbi(self.simulate, self.prior)

        # Set up posterior
        # hidden_features and num_transforms are sbi-specific (do not depend on any other params)
        neural_posterior = utils.posterior_nn(
            model="maf",
            embedding_net=feature_model,
            hidden_features=10,
            num_transforms=2,
        )

        # Set up an SBI's optimizer (default = SNPE_C)
        self.inference = SNPE(prior=self.prior, density_estimator=neural_posterior)

        proposal = self.prior
        for _ in range(num_epochs):
            # Train the feature model

            # theta = simulated parameters, shape = (num_simulations, number of optimization params)
            # x = samples obtained with these parameters, shape = (num_simulations, number of CI * 1024)
            # https://www.mackelab.org/sbi/reference/#sbi.inference.base.simulate_for_sbi
            theta, x = simulate_for_sbi(
                self.simulator,
                self.prior,
                num_simulations=num_simulations,
                num_workers=workers,
                show_progress_bar=verbose,
            )
            # Reshape x for convenience
            x = x.reshape((-1, self.num_current_injections, 1024))

            # Train the model for each value of current injections (sbi only works with shapes (1, ...))
            # The model is saved after each current injections, so we basically iterate over a batch
            for i in range(self.num_current_injections):
                density_estimator = self.inference.append_simulations(theta, x[:, i, :])
                # Note: sbi' API has resume_training parameter in .train(...) which defaults to False,
                # and the method breaks if it's set to True. However, we checked the model's weights
                # are saved after each iteration even with resume_training = False, so the code
                # works as intended
                density_estimator.train(show_train_summary=verbose)

            # Build posterior
            proposal = self.inference.build_posterior()
            self.posterior.append(proposal)

            # Set proposal to be the most probable posterior sample
            # x should be of the same shape as current injections, so we again iterate over the batch
            for i in range(self.num_current_injections):
                samples = self.posterior[-1].sample(
                    (1000,), x=self.observed_data[i, :], show_progress_bars=verbose
                )
                log_prob = self.posterior[-1].log_prob(
                    samples, x=self.observed_data[i, :], norm_posterior=False
                )
                proposal = samples[np.argmax(log_prob)]

        # Sample from the posterior
        out = []
        for i in range(self.num_current_injections):
            out.append(
                self.posterior[-1]
                .sample(
                    (num_samples,),
                    x=self.observed_data[i, :],
                    show_progress_bars=verbose,
                )
                .numpy()
            )

        return np.array(out)

    def simulate(self, *args, **kwargs) -> torch.Tensor:
        """
        Simulate data for SBI.

        Returns:
        ----------
        outs: torch.Tensor(shape = (num_current_injections, 1024))
            Simulated data.
        """

        simulated_data = []

        for inj in self.current_injections:
            # Set simulation parameters
            steps_per_ms = 10
            h.steps_per_ms = steps_per_ms
            h.dt = 1 / steps_per_ms
            h.tstop = self.tstop
            h.v_init = self.v_init

            self.i_clamp.dur = self.i_clamp_dur
            self.i_clamp.amp = inj
            self.i_clamp.delay = self.i_clamp_delay

            self.cell.set_parameters(self.parameters, args[0])
            self.cell.set_parameters(list(kwargs.keys()), list(kwargs.values()))

            # Run the simulation with the given parameters
            h.run()

            voltage = self.cell.mem_potential.as_numpy()
            voltage = resample(voltage, self.observed_data.shape[1])
            simulated_data.append(voltage)

        out = torch.tensor(np.array(simulated_data)).float()

        return out

    def optimize_with_segregation(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        parameter_inds: list,
        voltage_bounds: list,
        num_epochs: int = 1,
        num_simulations: int = 100,
        num_samples: int = 10,
        workers: int = 1,
        verbose=False,
    ) -> np.array:
        """
        Optimize using the segregation scheme from Alturki2016.

        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        parameter_inds: list
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        num_groves: int = 100
            Number of groves to average across. Equivalent to the number of epochs.

        num_prediction_rounds: int = 10
            Number of prediction trials.

        tree_max_depth: int = 5
            Max depth of each tree.

        Returns:
        ----------
        estimates: ndarray, shape = (number of parameters, )
            Parameter estimates.
        """
        # Get data to estimate on and set parameters to the user-defined values;
        # Save original parameters for a future reset
        cut_observed_data, original_param_set = self.prepare_for_segregation(
            observed_data, parameter_inds, voltage_bounds
        )

        # Optimize
        out = self.optimize(
            feature_model=feature_model,
            observed_data=cut_observed_data,
            num_epochs=num_epochs,
            num_simulations=num_simulations,
            num_samples=num_samples,
            workers=workers,
            verbose=verbose,
        )

        # Reset parameters back to the full sets
        self.parameters, self.lows, self.highs, self.num_parameters = original_param_set

        return out


class NaiveLinearOptimizer(ACTOptimizer):
    """
    At each epoch, Naive Linear Optimizer generates random parameter samples from the uniform distribution
    on given [lows, highs] bounds and runs NEURON on these samples for each value of current injection amplitude.
    NEURON's voltage output is then used as training data for the predictive model, which tries to find parameters
    that could produce such output.

    The predictive model consists of two linear layers separated by a ReLU activation. The output of the second linear layer
    is passed through a sigmoid function and multiplied by (highs - lows) + lows to ensure predictions lie in the given bounds.
    The network is optimized with the ADAM optimizer to minimize MSE loss.

    The pipeline does not require any external data and uses observed_data only for predictions. It also does not support feature models.

    The optimizer outputs parameter estimates for each current injection amplidtude. Ideally, constant parameter estimates should have
    the same value across all amplitudes, but if that is not the case, the estimates can be treated as different potential solutions.

    The optimizer is known to get stuck in local minima, so multiple re-runs are advised.
    """

    def __init__(self, config_file: str) -> None:
        """
        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """
        super().__init__(config_file)

    def optimize(
        self,
        observed_data: torch.Tensor,
        num_epochs: int = 100,
        lr: float = 1e-3,
        verbose: bool = False,
        return_loss_history: bool = False,
    ) -> np.array:
        """
        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for. They will only be used for predictions.

        num_epochs: int = 100
            Number of epochs (rounds) to run.

        lr: float = 1e-3
            Learning rate for the ADAM optimizer.

        verbose: bool = False
            Whether to print loss at each epoch. The loss is averaged across all current injections.

        return_loss_history: bool = False
            Whether to return loss history together with parameter esitmates.

        Returns:
        ----------
        estimates: ndarray, shape = (num_current_injections, number of parameters)
            Parameter estimates.

        loss_history: list
            If return_loss_history = True, list of loss values at each epoch.
        """

        self.model = torch.nn.Sequential(
            torch.nn.Linear(observed_data.shape[1], 16),
            torch.nn.ReLU(),
            torch.nn.Linear(16, self.num_parameters),
            torch.nn.Sigmoid(),
        )

        # Scalers to ensure parameter predictions are within range
        # Ref: https://stackoverflow.com/questions/73071399/how-to-bound-the-output-of-a-layer-in-pytorch
        sigmoid_mins = torch.tensor(self.lows)
        sigmoid_maxs = torch.tensor(self.highs)

        optim = torch.optim.Adam(self.model.parameters(), lr=lr)

        self.model.train(True)
        loss_history = []
        for ne in range(num_epochs):
            # Generate random parameter samples in the specified bounds
            param_samples = np.random.uniform(low=self.lows, high=self.highs)

            # Simulate a batch with generated parameter samples
            simulated_data = self.simulate(param_samples)

            # Resample to match the length of target data
            resampled_data = []
            for i in range(self.num_current_injections):
                resampled_data.append(
                    resample(x=simulated_data[i], num=observed_data.shape[1])
                )

            resampled_data = torch.tensor(np.array(resampled_data)).float()

            # Do standard optimization
            losses = []
            for i in range(self.num_current_injections):
                optim.zero_grad()
                pred = (
                    self.model(resampled_data[i]) * (sigmoid_maxs - sigmoid_mins)
                    + sigmoid_mins
                )

                loss = torch.nn.functional.mse_loss(
                    pred, torch.tensor(param_samples).float()
                )
                losses.append(loss.detach().numpy())

                loss.backward()
                optim.step()

            if verbose:
                print(f"epoch {ne}: mean_train_loss = {np.mean(losses)}")

            loss_history.append(np.mean(losses))

        # Get predictions
        self.model.eval()
        outs = []
        for i in range(self.num_current_injections):
            out = (
                self.model(observed_data[i]) * (sigmoid_maxs - sigmoid_mins)
                + sigmoid_mins
            )
            outs.append(out.detach().numpy())

        if return_loss_history:
            return (
                np.array(outs).reshape((self.num_current_injections, -1)),
                loss_history,
            )
        else:
            return np.array(outs).reshape((self.num_current_injections, -1))

    def optimize_with_segregation(
        self,
        observed_data: torch.Tensor,
        parameter_inds: list,
        voltage_bounds: list,
        num_epochs: int = 100,
        lr: float = 1e-3,
        verbose: bool = False,
        return_loss_history: bool = False,
    ) -> np.array:
        """
        Optimize using the segregation scheme from Alturki2016.

        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for. They will only be used for predictions.

        parameter_inds: list
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        num_epochs: int = 100
            Number of epochs (rounds) to run.

        lr: float = 1e-3
            Learning rate for the ADAM optimizer.

        verbose: bool = False
            Whether to print loss at each epoch. The loss is averaged across all current injections.

        return_loss_history: bool = False
            Whether to return loss history together with parameter esitmates.

        Returns:
        ----------
        estimates: ndarray, shape = (num_current_injections, number of parameters)
            Parameter estimates.

        loss_history: list
            If return_loss_history = True, list of loss values at each epoch.
        """
        # Get data to estimate on and set parameters to the user-defined values;
        # Save original parameters for a future reset
        cut_observed_data, original_param_set = self.prepare_for_segregation(
            observed_data, parameter_inds, voltage_bounds
        )

        # Optimize
        out = self.optimize(
            cut_observed_data, num_epochs, lr, verbose, return_loss_history
        )

        # Reset parameters back to the full sets
        self.parameters, self.lows, self.highs, self.num_parameters = original_param_set

        return out


class RandomSearchLinearOptimizer(ACTOptimizer):
    """
    At each epoch, Random Search Linear Optimizer generates a random parameter sample from the uniform on [lows, highs] distribution
    and runs NEURON with the generated sample. NEURON's voltage output is passed through a feature model and a linear layer with
    ReLU activation to obtain predictions which are then optimized to match the observed data. Both the feature and linear models
    are trained simultaneoulsy with the SGD optimizer and MSE loss.

    Predictions for the observed data are made with random search. At each round, a random parameter sample is generated and passed
    through NEURON and the trained model, and the loss value is saved to the list. The algorithm outputs the parameter sample with
    the lowest loss value across prediction rounds.

    Due to the nature of the algorithm, training loss values on general data will probably be large and unstable.
    """

    def __init__(self, config_file: str) -> None:
        """
        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """
        super().__init__(config_file)

    def optimize(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        num_summary_features: int,
        num_epochs: int = 100,
        num_prediction_rounds: int = 100,
        lr: float = 1e-3,
        verbose: bool = False,
        return_loss_history: bool = False,
    ) -> np.array:
        """
        Optimize using a linear model.

        Parameters:
        ----------
        feature_model: torch.nn.Module
            Model that generates summary features. Accepts tensors of shape (num_current_injections, len_voltage_trace).
            Returns tensors of shape (num_current_injections, num_summary_features).

        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        num_summary_features: int
            Number of summary features which a feature model outputs.

        num_epochs: int = 10
            Number of training epochs to run.

        num_prediction_rounds: int = 10
            Number of prediction trials.

        verbose: bool = False
            Whether to print training loss at each epoch. The loss is averaged across current injections.

        return_loss_history: bool = False
            Whether to return loss history together with parameter esitmates.

        Returns:
        ----------
        estimates: ndarray, shape = (number of parameters, )
            Parameter estimates.

        loss_history: list
            If return_loss_history = True, list of loss values at each epoch.
        """

        self.model = torch.nn.Sequential(
            feature_model,
            torch.nn.Linear(num_summary_features, observed_data.shape[1]),
            torch.nn.ReLU(),
        )

        optim = torch.optim.SGD(self.model.parameters(), lr=lr)

        self.model.train(True)
        loss_history = []
        for ne in range(num_epochs):
            # Generate a random parameter sample in the specified bounds
            param_samples = np.random.uniform(low=self.lows, high=self.highs)

            # Simulate a batch with the generated parameter sample
            simulated_data = self.simulate(param_samples)

            # Resample to match the length of target data
            resampled_data = []
            for i in range(self.num_current_injections):
                resampled_data.append(
                    resample(x=simulated_data[i], num=observed_data.shape[1])
                )

            resampled_data = torch.tensor(np.array(resampled_data)).float()

            # Do standard optimization
            losses = []
            for i in range(self.num_current_injections):
                optim.zero_grad()
                pred = self.model(resampled_data[i])

                loss = torch.nn.functional.mse_loss(
                    pred, observed_data[i].reshape((1, -1))
                )
                losses.append(loss.detach().numpy())

                loss.backward()
                optim.step()

            if verbose:
                print(f"epoch {ne}: mean_train_loss = {np.mean(losses)}")
            if return_loss_history:
                loss_history.append(np.mean(losses))

        # Predict with random search
        predicts = []
        losses = []

        self.model.eval()
        with torch.no_grad():
            for _ in range(num_prediction_rounds):
                # Generate random parameter samples in the specified bounds
                param_samples = np.random.uniform(low=self.lows, high=self.highs)

                # Simulate a batch with generated parameter samples
                simulated_data = self.simulate(param_samples)

                # Resample to match the length of target data
                resampled_data = []
                for i in range(self.num_current_injections):
                    resampled_data.append(
                        resample(x=simulated_data[i], num=observed_data.shape[1])
                    )

                resampled_data = torch.tensor(np.array(resampled_data)).float()

                candidate_loss = []
                for i in range(self.num_current_injections):
                    pred = self.model(resampled_data[i])
                    loss = torch.nn.functional.mse_loss(
                        pred, observed_data[i].reshape((1, -1))
                    )
                    candidate_loss.append(loss.detach().numpy())

                predicts.append(param_samples)
                losses.append(np.mean(candidate_loss))

        final_prediction = np.array(predicts[np.argmin(losses)]).flatten()

        if return_loss_history:
            return final_prediction, loss_history
        else:
            return final_prediction

    def optimize_with_segregation(
        self,
        feature_model: torch.nn.Module,
        observed_data: torch.Tensor,
        num_summary_features: int,
        parameter_inds: list,
        voltage_bounds: list,
        num_epochs: int = 100,
        num_prediction_rounds: int = 100,
        lr: float = 1e-3,
        verbose: bool = False,
        return_loss_history: bool = False,
    ) -> np.array:
        """
        Optimize using the segregation scheme from Alturki2016.

        Parameters:
        ----------
        feature_model: torch.nn.Module
            Model that generates summary features. Accepts tensors of shape (num_current_injections, len_voltage_trace).
            Returns tensors of shape (num_current_injections, num_summary_features).

        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        num_summary_features: int
            Number of summary features which a feature model outputs.

        parameter_inds: list
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        num_epochs: int = 10
            Number of training epochs to run.

        num_prediction_rounds: int = 10
            Number of prediction trials.

        verbose: bool = False
            Whether to print training loss at each epoch. The loss is averaged across current injections.

        return_loss_history: bool = False
            Whether to return loss history together with parameter esitmates.

        Returns:
        ----------
        estimates: ndarray, shape = (number of parameters, )
            Parameter estimates.

        loss_history: list
            If return_loss_history = True, list of loss values at each epoch.
        """
        # Get data to estimate on and set parameters to the user-defined values;
        # Save original parameters for a future reset
        cut_observed_data, original_param_set = self.prepare_for_segregation(
            observed_data, parameter_inds, voltage_bounds
        )

        # Optimize
        out = self.optimize(
            feature_model=feature_model,
            observed_data=cut_observed_data,
            num_summary_features=num_summary_features,
            num_epochs=num_epochs,
            num_prediction_rounds=num_prediction_rounds,
            lr=lr,
            verbose=verbose,
            return_loss_history=return_loss_history,
        )

        # Reset parameters back to the full sets
        self.parameters, self.lows, self.highs, self.num_parameters = original_param_set

        return out


class RandomSearchTreeOptimizer(ACTOptimizer):
    """
    At each epoch, Random Search Tree Optimizer generates a random parameter sample from the uniform on [lows, highs] distribution
    and runs NEURON with the generated sample. NEURON's voltage output for each current injection value is then used to train an
    independent decision tree to predict the voltage trace for the same current injection value from the observed data. Trees for
    a single parameter sample are united into groves (lists of fitted decision trees). The final model is a list of groves.

    Predictions for the observed data are made with random search. At each round, a random parameter sample is generated and passed
    through NEURON and then averaged across all groves for each value of current injection. The algorithm outputs the parameter sample with
    the lowest loss value across prediction rounds.

    The tree model is provided by scikit-learn, so all methods and attributes from
    https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html can be used.
    """

    def __init__(self, config_file: str) -> None:
        """
        Parameters:
        ----------
        config_file: str
            Path to the .json configuration file with simulation parameters' values.
        """
        super().__init__(config_file)

    def optimize(
        self,
        observed_data: torch.Tensor,
        num_groves: int = 100,
        num_prediction_rounds: int = 10,
        tree_max_depth: int = 5,
    ) -> np.array:
        """
        Optimize using a decision tree.

        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        num_groves: int = 100
            Number of groves to average across. Equivalent to the number of epochs.

        num_prediction_rounds: int = 10
            Number of prediction trials.

        tree_max_depth: int = 5
            Max depth of each tree.

        Returns:
        ----------
        estimates: ndarray, shape = (number of parameters, )
            Parameter estimates.
        """

        self.model = []

        for _ in range(num_groves):
            # Generate a random parameter sample in the specified bounds
            param_samples = np.random.uniform(low=self.lows, high=self.highs)

            # Simulate a batch with the generated parameter sample
            simulated_data = self.simulate(param_samples)

            # Resample to match the length of target data
            resampled_data = []
            for i in range(self.num_current_injections):
                resampled_data.append(
                    resample(x=simulated_data[i], num=observed_data.shape[1])
                )

            resampled_data = np.array(resampled_data)

            # Train a tree for each value of current injection
            grove = []
            for i in range(self.num_current_injections):
                tree = DecisionTreeRegressor(max_depth=tree_max_depth)
                tree.fit(resampled_data[i].reshape((-1, 1)), observed_data[i])
                grove.append(tree)

            self.model.append(grove)

        # Predict with random search
        predicts = []
        losses = []

        for _ in range(num_prediction_rounds):
            # Generate random parameter samples in the specified bounds
            param_samples = np.random.uniform(low=self.lows, high=self.highs)

            # Simulate a batch with generated parameter samples
            simulated_data = self.simulate(param_samples)

            # Resample to match the length of target data
            resampled_data = []
            for i in range(self.num_current_injections):
                resampled_data.append(
                    resample(x=simulated_data[i], num=observed_data.shape[1])
                )

            resampled_data = np.array(resampled_data)

            # Final prediction is the average prediction across all trees for each value of current injection
            candidate_loss = []
            for i in range(self.num_current_injections):
                pred_for_ci = 0
                for j in range(num_groves):
                    pred_for_ci += self.model[j][i].predict(
                        resampled_data[i].reshape((-1, 1))
                    )
                pred_for_ci = pred_for_ci / num_groves

                loss = mean_squared_error(
                    observed_data[i].detach().numpy(), pred_for_ci
                )
                candidate_loss.append(loss)

            predicts.append(param_samples)
            losses.append(np.mean(candidate_loss))

        final_prediction = np.array(predicts[np.argmin(losses)]).flatten()

        return final_prediction

    def optimize_with_segregation(
        self,
        observed_data: torch.Tensor,
        parameter_inds: list,
        voltage_bounds: list,
        num_groves: int = 100,
        num_prediction_rounds: int = 10,
        tree_max_depth: int = 5,
    ) -> np.array:
        """
        Optimize using the segregation scheme from Alturki2016.

        Parameters:
        ----------
        observed_data: torch.Tensor(shape = (num_current_injections, len_voltage_trace))
            Target values to optimize for.

        parameter_inds: list
            Indexes of parameters to be optimized for.

        voltage_bounds: list[int]
            Lower and upper bound of the region which will be cut from the voltage trace.

        num_groves: int = 100
            Number of groves to average across. Equivalent to the number of epochs.

        num_prediction_rounds: int = 10
            Number of prediction trials.

        tree_max_depth: int = 5
            Max depth of each tree.

        Returns:
        ----------
        estimates: ndarray, shape = (number of parameters, )
            Parameter estimates.
        """
        # Get data to estimate on and set parameters to the user-defined values;
        # Save original parameters for a future reset
        cut_observed_data, original_param_set = self.prepare_for_segregation(
            observed_data, parameter_inds, voltage_bounds
        )

        # Optimize
        out = self.optimize(
            observed_data=cut_observed_data,
            num_groves=num_groves,
            num_prediction_rounds=num_prediction_rounds,
            tree_max_depth=tree_max_depth,
        )

        # Reset parameters back to the full sets
        self.parameters, self.lows, self.highs, self.num_parameters = original_param_set

        return out



import numpy as np
import torch
from scipy import stats

# Information to calculate 1D CNN output size and maxpool output size.
# https://towardsdatascience.com/pytorch-basics-how-to-train-your-neural-net-intro-to-cnn-26a14c2ea29
# Maxpool layer uses same resource but kernel size and stride are set to be the same unless otherwise specified.
# Padding set to 0 by default.


class DefaultSummaryModel(torch.nn.Module):
    def __init__(
        self,
        num_linear_layer_input_features: int,
        num_summary_features: int = 8,
        use_statistics: bool = True,
    ) -> None:
        """
        CNN to get summary features from current injections data.

        Parameters:
        ----------

        num_current_injections: int
            Number of current injections. Used as the number of input channels for the network.

        num_summary_features: int
            Number of summary features. Equal to the number of ouput dimensions of the network.

        hybrid: bool
            If true, concatenate the network's output with summary statistics of the input data (mean, variance and skew).
        """
        super().__init__()

        # Save for forward pass
        self.use_statistics = use_statistics
        self.num_linear_layer_input_features = num_linear_layer_input_features

        # Each layer is initialized individually to implement skip-connections
        # Dimensionalities a -> b below are arbitrary to keep track of changes
        self.conv1 = torch.nn.Conv1d(
            in_channels=1, out_channels=1, kernel_size=9, padding="same"
        )  # 1024 -> 1024
        self.pool1 = torch.nn.MaxPool1d(kernel_size=2, stride=2)  # 1024 -> 512
        self.conv2 = torch.nn.Conv1d(
            in_channels=1, out_channels=1, kernel_size=7, padding="same"
        )  # 512 -> 512
        self.pool2 = torch.nn.MaxPool1d(kernel_size=2, stride=2)  # 512 -> 256
        self.conv3 = torch.nn.Conv1d(
            in_channels=1, out_channels=1, kernel_size=5, padding="same"
        )  # 256 -> 256
        self.pool3 = torch.nn.MaxPool1d(kernel_size=2, stride=2)  # 256 -> 128
        self.conv4 = torch.nn.Conv1d(
            in_channels=1, out_channels=1, kernel_size=3, padding="same"
        )  # 128 -> 128
        self.pool4 = torch.nn.MaxPool1d(kernel_size=2, stride=2)  # 128 -> 64
        self.linear = torch.nn.Linear(
            in_features=num_linear_layer_input_features,
            out_features=num_summary_features,
        )

        print(
            f"Total number of summary features: {num_summary_features + use_statistics * 3}"
        )

    def forward(self, X):
        X = X.reshape((-1, 1, X.shape[0]))  # e.g., (batch_size, 1, 1024)

        out = self.pool1(torch.nn.functional.relu(self.conv1(X)))

        sk_con = self.pool2(out)
        out = self.pool2(torch.nn.functional.relu(self.conv2(out)))
        out = out + sk_con

        sk_con = self.pool3(out)
        out = self.pool3(torch.nn.functional.relu(self.conv3(out)))
        out = out + sk_con

        sk_con = self.pool4(out)
        out = self.pool4(torch.nn.functional.relu(self.conv4(out)))

        out = out.view(1, -1)
        out = self.linear(out)

        if self.use_statistics:
            out = torch.cat((out, self.compute_statistics(X)), 1)

        return out

    def compute_statistics(self, X: torch.tensor) -> torch.tensor:
        """
        Calculates mean, variance and skew of input data's columns.

        Parameters:
        ----------
        X: torch.tensor
            Data to compute on.

        Returns:
        ----------
        out: torch.tensor
            Tensor with summary statistics.
        """
        data = X.numpy()
        mean = np.mean(data, -1)
        variance = np.std(data, -1)
        skew = stats.skew(data, -1)

        return torch.from_numpy(np.concatenate((mean, variance, skew), -1))



import numpy as np

from act.act_types import SimulationConfig

pospischilsPY: SimulationConfig = {
    "cell": {
        "hoc_file": "../data/Pospischil/sPy/template.hoc",
        "modfiles_folder": "../data/Pospischil/sPy/seg_modfiles",
        "name": "sPY",
    },
    "simulation_parameters": {
        "h_v_init": -67.0,  # (mV)
        "h_tstop": 2000,  # (ms)
        "h_i_delay": 500,  # (ms)
        "h_i_dur": 1500,  # (ms)
        "h_dt": 0.025,
    },
    "optimization_parameters": {
        "amps": list(np.arange(-2, 10, 0.1)),
        "params": [
            {"channel": "g_pas", "low": 5.0e-06, "high": 2.0e-05},
            {"channel": "gnabar_hh2", "low": 2.5e-02, "high": 1.0e-01},
            {"channel": "gkbar_hh2", "low": 2.5e-03, "high": 1.0e-02},
            {"channel": "gkbar_im", "low": 1.5e-05, "high": 6.0e-05},
        ],
        "target_V": None,  # Target voltages
        "target_params": [0.0001, 0.05, 0.005, 7e-5],
        "num_repeats": 3,
        "num_amps_to_match": 12,
        "num_epochs": 5000,
    },
    "summary_features": {
        "spike_threshold": 20,  # (mV)
        # Target-sim match conditions (max abs diff between sim and target)
        "mc_num_spikes": 1,
        "mc_interspike_time": 200,  # (ms)
        "mc_min_v": 1,  # (mV)
        "mc_mean_v": 2,  # (mV)
        "mc_max_v": 1,  # (mV)
    },
    "segregation": [
        {
            "params": ["g_pas"],
            "voltage": [-100, -65],
            "time": [0, 500],
        },
        {
            "params": ["gnabar_hh2", "gkbar_hh2", "gkbar_im"],
            "voltage": [-65, 100],
            "time": [0, 2000],
        },
    ],
    "output": {"folder": "output_Pospischil_sPY", "produce_plots": True},
    "run_mode": "segregated",  # "original", "segregated"
}


pospischilsPYr: SimulationConfig = {
    "cell": {
        "hoc_file": "../data/Pospischil/sPyr/template.hoc",
        "modfiles_folder": "../data/Pospischil/sPyr/seg_modfiles",
        "name": "sPYr",
    },
    "simulation_parameters": {
        "h_v_init": -67.0,  # (mV)
        "h_tstop": 2000,  # (ms)
        "h_i_delay": 500,  # (ms)
        "h_i_dur": 1500,  # (ms)
        "h_dt": 0.025,
    },
    "optimization_parameters": {
        "amps": list(np.arange(-2, 10, 0.1)),
        "params": [
            {"channel": "g_pas", "low": 5.0e-06, "high": 2.0e-05},
            {"channel": "gnabar_hh2", "low": 2.5e-02, "high": 1.0e-01},
            {"channel": "gkbar_hh2", "low": 2.5e-03, "high": 1.0e-02},
            {"channel": "gkbar_im", "low": 1.5e-05, "high": 6.0e-05},
            {"channel": "gcabar_it", "low": 5.0e-04, "high": 2.0e-03},
        ],
        "target_V": None,  # Target voltages
        "target_params": [1e-5, 0.05, 0.005, 3e-5, 0.001],
        "num_repeats": 3,
        "num_amps_to_match": 12,
        "num_epochs": 5000,
    },
    "summary_features": {
        "spike_threshold": 20,  # (mV)
        # Target-sim match conditions (max abs diff between sim and target)
        "mc_num_spikes": 1,
        "mc_interspike_time": 200,  # (ms)
        "mc_min_v": 1,  # (mV)
        "mc_mean_v": 2,  # (mV)
        "mc_max_v": 1,  # (mV)
    },
    "segregation": [
        {
            "params": ["g_pas"],
            "voltage": [-100, -65],
            "time": [0, 500],
        },
        {
            "params": ["gnabar_hh2", "gkbar_hh2", "gkbar_im", "gcabar_it"],
            "voltage": [-65, 100],
            "time": [0, 2000],
        },
    ],
    "output": {"folder": "output_Pospischil_sPYr", "produce_plots": True},
    "run_mode": "segregated",  # "original", "segregated"
}

pospischilsPYr_passive: SimulationConfig = {
    "cell": {
        "hoc_file": "../data/Pospischil/sPyr/template.hoc",
        "modfiles_folder": "../data/Pospischil/sPyr/seg_modfiles",
        "name": "sPYr",
        "passive_properties": {
            "v_rest": -60,
            "r_in": 393.45,
            "tau": 117.425,
            "leak_conductance_variable": "g_pas",  # eg: g_leak
            "leak_reversal_variable": "e_pas",  # eg: e_leak
        },
    },
    "simulation_parameters": {
        "h_v_init": -60.0,  # (mV)
        "h_tstop": 2000,  # (ms)
        "h_i_delay": 500,  # (ms)
        "h_i_dur": 1500,  # (ms)
        "h_dt": 0.025,
    },
    "optimization_parameters": {
        "amps": list(np.arange(-2, 10, 0.1)),
        "params": [
            {"channel": "gnabar_hh2", "low": 2.5e-02, "high": 1.0e-01},
            {"channel": "gkbar_hh2", "low": 2.5e-03, "high": 1.0e-02},
            {"channel": "gkbar_im", "low": 1.5e-05, "high": 6.0e-05},
            {"channel": "gcabar_it", "low": 5.0e-04, "high": 2.0e-03},
        ],
        "target_V": None,  # Target voltages
        "target_params": [0.05, 0.005, 3e-5, 0.001],
        "num_repeats": 3,
        "num_amps_to_match": 12,
        "num_epochs": 5000,
    },
    "summary_features": {
        "spike_threshold": 20,  # (mV)
        # Target-sim match conditions (max abs diff between sim and target)
        "mc_num_spikes": 1,
        "mc_interspike_time": 200,  # (ms)
        "mc_min_v": 1,  # (mV)
        "mc_mean_v": 2,  # (mV)
        "mc_max_v": 1,  # (mV)
    },
    "segregation": [
        {
            "params": ["gnabar_hh2", "gkbar_hh2", "gkbar_im", "gcabar_it"],
            "voltage": [-65, 100],
            "time": [0, 2000],
        },
    ],
    "output": {"folder": "output_Pospischil_sPYr_p", "produce_plots": True},
    "run_mode": "segregated",  # "original", "segregated"
}

LA_A_seg = {
    "cell": {
        "hoc_file": "../data/LA/A/template.hoc",
        "modfiles_folder": "../data/LA/A/seg_modfiles_modeldb",
        "name": "Cell_A",
        "passive_properties": {
            "v_rest": -69.17387,
            "r_in": 141,
            "tau": 30.88,
            "leak_conductance_variable": "glbar_leak",  # eg: g_leak
            "leak_reversal_variable": "el_leak",  # eg: e_leak
        },
    },
    "simulation_parameters": {
        "h_v_init": -70.0,  # (mV)
        "h_tstop": 2000,  # (ms)
        "h_i_delay": 500,  # (ms)
        "h_i_dur": 1500,  # (ms)
        "h_dt": 0.025,
    },
    "optimization_parameters": {
        "amps": list(np.arange(-2, 3, 0.025)),
        "params": [
            # {"channel": "ghdbar_hd", "low": 1.15e-05, "high": 4.6e-05}, # hd, passive
            {
                "channel": "gbar_nap",
                "low": 0.000071,
                "high": 0.000284,
            },  # nap, lto and hto
            {"channel": "gbar_im", "low": 0.001, "high": 0.004},  # im, lto and hto
            {
                "channel": "gbar_na3",
                "low": 0.015,
                "high": 0.06,
            },  # na3, spiking/adaptation
            {
                "channel": "gbar_kdr",
                "low": 0.00075,
                "high": 0.003,
            },  # kdr, spiking/adaptation
            {
                "channel": "gcabar_cadyn",
                "low": 3e-5,
                "high": 1.2e-4,
            },  # cadyn, spiking/adaptation
            {
                "channel": "gsAHPbar_sAHP",
                "low": 0.0045,
                "high": 0.018,
            },  # sahp, spiking/adaptation
        ],
        "target_V": None,  # Target voltages
        "target_params": [
            0.000142,
            0.002,
            0.03,
            0.0015,
            6e-5,
            0.009,
        ],  # [2.3e-05, 0.000142, 0.002, 0.03, 0.0015, 6e-5, 0.009],
        "num_repeats": 3,
        "num_amps_to_match": 12,
        "num_epochs": 5000,
        "parametric_distribution": {  # sample the parameter space for training if n_slices is > 1
            "n_slices": 10,
            "simulations_per_amp": 3,  # randomly select n permutations per amp to train on
        },
    },
    "summary_features": {
        "spike_threshold": 20,  # (mV)
        # Target-sim match conditions (max abs diff between sim and target)
        "mc_num_spikes": 1,
        "mc_interspike_time": 200,  # (ms)
        "mc_min_v": 1,  # (mV)
        "mc_mean_v": 2,  # (mV)
        "mc_max_v": 1,  # (mV)
    },
    "segregation": [
        # { # passive
        #    "params": ["ghdbar_hd"],
        #    "voltage": [-80, -67.5],
        # },
        {  # lto
            "params": ["gbar_nap", "gbar_im"],
            "voltage": [-100, 100],  # [-67.5, 100],  # [-67.5, -57.5],
        },
        {  # spking / adaptation
            "params": ["gbar_na3", "gbar_kdr", "gcabar_cadyn", "gsAHPbar_sAHP"],
            "voltage": [-100, 100],  # [-57.5, 100],  # [-57.5, 0],
        },
        {  # hto
            "params": ["gbar_nap", "gbar_im"],
            "voltage": [-100, 100],  # [-40, 100],  # [-40, -30],
        },
    ],
    "output": {
        "folder": "output_LA_A",
        "produce_plots": True,
        "target_label": "ModelDB Segregated",
        "simulated_label": "Model ACT",
    },
    "run_mode": "segregated",  # "original", "segregated"
}



import json
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch

from act.act_types import SimulationConfig
from act.metrics import correlation_score, mse_score
from act.optim import ACTOptimizer


def save_plot(
    amp: float,
    output_folder: str,
    simulated_data=None,
    target_V=None,
    output_file=None,
    simulated_label="Simulated",
    target_label="Target",
    dt=0.025,
):
    _, ax = plt.subplots(1, 1, figsize=(10, 10))
    title = f"I = {(amp * 1000):.0f} nA"
    if simulated_data is not None:
        times = np.arange(0, int(len(simulated_data.flatten()) * dt), dt)
        ax.plot(times, simulated_data.flatten(), label=simulated_label)
    if target_V is not None:
        times = np.arange(0, int(len(target_V.flatten()) * dt), dt)
        ax.plot(times, target_V.flatten(), label=target_label)
    ax.set_title(title)
    ax.set_xlabel("Timestamp (ms)")
    ax.set_ylabel("V (mV)")
    ax.legend()
    ax.grid()

    if not output_file:
        output_file = os.path.join(output_folder, f"{(amp * 1000):.0f}nA.png")
    else:
        output_file = os.path.join(output_folder, output_file)
    plt.savefig(output_file)
    plt.close()


def save_prediction_plots(
    target_V: torch.Tensor,
    amp: list,
    simulation_config: SimulationConfig,
    predicted_params_values: torch.Tensor,
    output_folder: str,
) -> np.ndarray:
    optim = ACTOptimizer(simulation_config=simulation_config)
    params = [
        p["channel"] for p in simulation_config["optimization_parameters"]["params"]
    ]
    simulated_data = optim.simulate(
        amp, params, predicted_params_values.detach().numpy()
    )
    simulated_data = optim.resample_voltage(
        V=simulated_data.reshape((1, -1)), num_obs=target_V.shape[1]
    )
    dt = simulation_config["simulation_parameters"]["h_dt"]
    simulated_label = simulation_config["output"].get("simulated_label", "Simulated")
    target_label = simulation_config["output"].get("target_label", "Target")
    save_plot(
        amp,
        output_folder,
        simulated_data,
        target_V,
        dt=dt,
        simulated_label=simulated_label,
        target_label=target_label,
    )

    return simulated_data


def save_mse_corr(
    target_V: torch.Tensor,
    simulation_config: SimulationConfig,
    predicted_params_values: torch.Tensor,
    output_folder: str,
) -> None:
    with open(os.path.join(output_folder, "metrics.csv"), "w") as file:
        file.write(f"amp,mse,corr\n")

    optim = ACTOptimizer(simulation_config=simulation_config)
    for ind, amp in enumerate(simulation_config["optimization_parameters"]["amps"]):
        params = [
            p["channel"] for p in simulation_config["optimization_parameters"]["params"]
        ]
        sim_data = optim.simulate(amp, params, predicted_params_values.detach().numpy())
        simulated_data = optim.resample_voltage(
            V=sim_data.reshape((1, -1)), num_obs=target_V.shape[1]
        )
        mse = mse_score(target_V[ind].reshape(-1, 1), simulated_data.reshape(-1, 1))
        corr = correlation_score(
            target_V[ind].reshape(1, -1), simulated_data.reshape(1, -1)
        )

        with open(os.path.join(output_folder, "metrics.csv"), "a") as file:
            file.write(f"{amp},{mse},{corr}\n")


def print_run_stats(config: SimulationConfig):
    output_folder = config["output"]["folder"]
    run_mode = config["run_mode"]
    target_params = config["optimization_parameters"].get("target_params")
    pred_passive_json_path = os.path.join(
        output_folder, run_mode, "pred_passive_properties.json"
    )
    target_passive_json_path = os.path.join(
        output_folder, run_mode, "target", "target_passive_properties.json"
    )

    metrics = pd.read_csv(os.path.join(output_folder, run_mode, "metrics.csv"))
    preds_df = pd.read_csv(
        os.path.join(output_folder, run_mode, "pred.csv"), index_col=0
    )
    pred_passive_json = None
    if os.path.isfile(pred_passive_json_path):
        with open(pred_passive_json_path, "r") as fp:
            pred_passive_json = json.load(fp)

    target_passive_json = None
    if os.path.isfile(target_passive_json_path):
        with open(target_passive_json_path, "r") as fp:
            target_passive_json = json.load(fp)

    preds = np.array(preds_df)
    print(output_folder, ":", run_mode)
    print(f"Med MSE: {metrics['mse'].median():.4f} ({metrics['mse'].std():.4f})")
    print(f"Med Corr: {metrics['corr'].median():.4f} ({metrics['corr'].std():.4f})")
    print()
    print("Predicted values:")
    print(preds_df)
    if target_params:
        print("Target values:")
        print(pd.DataFrame([target_params], columns=preds_df.columns))
        print("Error:")
        print(preds_df - target_params)
        print()
        print(f"Pred MAE: {np.mean(np.abs(target_params - preds)):.4f}")
    if pred_passive_json:
        print()
        simulated_label = config["output"].get("simulated_label", "Simulated")
        print(f"{simulated_label} Passive properties:")
        print(json.dumps(pred_passive_json, indent=2))
        print("----------\n")
    if target_passive_json:
        print()
        target_label = config["output"].get("target_label", "Target")
        print(f"{target_label} Passive properties:")
        print(json.dumps(target_passive_json, indent=2))
        print("----------\n")



"""
These types should be used to define the user supplied simulation config/config.

"""

from typing import List, TypedDict


class PassiveProperties(TypedDict):
    v_rest: float
    r_in: float
    tau: float
    leak_conductance_variable: str  # eg: g_leak
    leak_reversal_variable: str  # eg: e_leak


class Cell(TypedDict):
    hoc_file: str
    modfiles_folder: str
    name: str
    passive_properties: PassiveProperties  # optional if you want an analytical passive approach


class SimulationParameters(TypedDict):
    h_v_init: float
    h_tstop: int  # (ms)
    h_i_delay: int  # (ms)
    h_i_dur: int  # (ms)
    h_dt: float


class OptimizationParam(TypedDict):
    param: str
    low: float
    high: float


class ParametricDistribution(TypedDict):
    n_slices: int  # slice each variable min to max, into n equal slices
    simulations_per_amp: int  # each amp will be split equally


class OptimizationParameters(TypedDict):
    amps: List[float]
    params: List[OptimizationParam]
    target_V: List[List[float]]  # Target voltage
    target_params: List[List[float]]
    num_repeats: int
    num_amps_to_match: int
    num_epochs: int
    parametric_distribution: ParametricDistribution


class SummaryFeatures:
    spike_threshold: int  # (mV)

    # Target-sim match conditions (max abs diff between sim and target)
    mc_num_spikes: int
    mc_interspike_time: int  # (ms)
    mc_min_v: int  # (mV)
    mc_mean_v: int  # (mV)
    mc_max_v: int  # (mV)


class SegregationModule(TypedDict):
    params: List[str]
    voltage: List[int]
    time: List[int]


class Output(TypedDict):
    folder: str
    produce_plots: bool
    target_label: str
    simulated_label: str


class SimulationConfig(TypedDict):
    cell: Cell
    simulation_parameters: SimulationParameters
    optimization_parameters: OptimizationParameters
    summary_features: SummaryFeatures
    segregation: List[SegregationModule]
    output: Output

    run_mode: str  # "original", "segregated"



import numpy as np
import torch
from neuron import h
from scipy.signal import resample

from act.act_types import PassiveProperties, SimulationConfig
from act.cell_model import CellModel
from act.logger import ACTDummyLogger


class ACTOptimizer:
    def __init__(
        self,
        simulation_config: SimulationConfig,
        logger: object = None,
        reset_cell_params_to_lower_bounds_on_init: bool = True,
        set_passive_properties=True,
    ):
        self.config = simulation_config

        # Initialize standard run
        h.load_file("stdrun.hoc")

        # Initialize the cell
        self.cell = CellModel(
            hoc_file=self.config["cell"]["hoc_file"],
            cell_name=self.config["cell"]["name"],
        )
        if set_passive_properties:
            self.cell.set_passive_properties(
                simulation_config["cell"].get("passive_properties")
            )

        if reset_cell_params_to_lower_bounds_on_init:
            params = [
                p["channel"] for p in self.config["optimization_parameters"]["params"]
            ]
            lows = [p["low"] for p in self.config["optimization_parameters"]["params"]]
            self.cell.set_parameters(params, lows)

        # For convenience
        self.update_param_vars()

        # Set the logger
        self.logger = logger
        if self.logger is None:
            self.logger = ACTDummyLogger()

    def optimize(self, target_V: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def optimize_with_segregation(
        self, target_V: torch.Tensor, segregate_by: str
    ) -> torch.Tensor:
        raise NotImplementedError

    def cut_voltage_region(
        self, target_V: torch.Tensor, voltage_bounds: list
    ) -> torch.Tensor:
        cut_target_V = torch.zeros_like(target_V)

        for i in range(cut_target_V.shape[0]):
            trace = target_V[i]
            trace = trace[(trace >= voltage_bounds[0]) & (trace < voltage_bounds[1])]
            cut_target_V[i] = torch.tensor(
                resample(x=trace, num=target_V.shape[1])
            ).float()

        return cut_target_V

    def cut_time_region(
        self, target_V: torch.Tensor, time_bounds: tuple
    ) -> torch.Tensor:
        cut_target_V = torch.zeros_like(target_V)

        for i in range(cut_target_V.shape[0]):
            trace = target_V[i, time_bounds[0] : time_bounds[1]]
            cut_target_V[i] = torch.tensor(
                resample(x=trace, num=target_V.shape[1])
            ).float()

        return cut_target_V

    def simulate(
        self,
        amp: float,
        parameter_names: list,
        parameter_values: np.ndarray,
        i_dur: float = 0,
        i_delay: float = 0,
    ) -> torch.Tensor:
        h.dt = self.config["simulation_parameters"]["h_dt"]
        h.tstop = self.config["simulation_parameters"]["h_tstop"]
        h.v_init = self.config["simulation_parameters"]["h_v_init"]

        self.cell.set_parameters(parameter_names, parameter_values)
        if not i_dur:
            i_dur = self.config["simulation_parameters"]["h_i_dur"]
        if not i_delay:
            i_delay = self.config["simulation_parameters"]["h_i_delay"]
        self.cell.apply_current_injection(
            amp,
            i_dur,
            i_delay,
        )

        h.run()

        return torch.Tensor(self.cell.Vm.as_numpy()[:-1])

    def calculate_passive_properties(
        self, parameter_names: list, parameter_values: np.ndarray
    ) -> (PassiveProperties, np.ndarray):
        """
        Run simulations to determine passive properties for a given cell parameter set
        """
        gleak_var = (
            self.config.get("cell", {})
            .get("passive_properties", {})
            .get("leak_conductance_variable")
        )
        eleak_var = (
            self.config.get("cell", {})
            .get("passive_properties", {})
            .get("leak_reversal_variable")
        )
        props: PassiveProperties = {
            "leak_conductance_variable": gleak_var,
            "leak_reversal_variable": eleak_var,
            "r_in": 0,
            "tau": 0,
            "v_rest": 0,
        }
        tstart = 500
        passive_delay = 500
        passive_amp = -100 / 1e3
        passive_duration = 1000

        passive_tensor = self.simulate(
            passive_amp,
            parameter_names,
            parameter_values,
            i_delay=tstart,
            i_dur=passive_duration,
        )
        passive_vec = passive_tensor.detach().numpy()

        index_v_rest = int(((1000 / h.dt) / 1000 * tstart))
        index_v_final = int(((1000 / h.dt) / 1000 * (tstart + passive_delay)))

        # v rest
        v_rest = passive_vec[index_v_rest]
        v_rest_time = index_v_rest / (1 / h.dt)

        # tau/r_in
        passive_v_final = passive_vec[index_v_final]
        v_final_time = index_v_final / (1 / h.dt)

        v_diff = v_rest - passive_v_final

        v_t_const = v_rest - (v_diff * 0.632)
        # Find index of first occurance where the voltage is less than the time constant for tau
        # index_v_tau = list(filter(lambda i: i < v_t_const, cfir_widget.passive_vec))[0]
        index_v_tau = next(
            x
            for x, val in enumerate(list(passive_vec[index_v_rest:]))
            if val < v_t_const
        )
        time_tau = index_v_tau / ((1000 / h.dt) / 1000)
        tau = time_tau  # / 1000 (in ms)
        r_in = (v_diff) / (0 - passive_amp)  # * 1e6  # MegaOhms -> Ohms

        props["v_rest"] = float(v_rest)
        props["tau"] = float(tau)
        props["r_in"] = float(r_in)

        return props, passive_vec

    def resample_voltage(self, V: torch.Tensor, num_obs: int) -> torch.Tensor:
        resampled_data = []
        for i in range(V.shape[0]):
            resampled_data.append(resample(x=V[i], num=num_obs))
        resampled_data = torch.tensor(np.array(resampled_data)).float()
        return resampled_data

    def update_param_vars(self) -> None:
        self.num_ampl = len(self.config["optimization_parameters"]["amps"])
        self.num_params = len(self.config["optimization_parameters"]["params"])


class GeneralACTOptimizer(ACTOptimizer):
    def __init__(
        self,
        simulation_config: SimulationConfig,
        logger: object = None,
        reset_cell_params_to_lower_bounds_on_init: bool = True,
    ):
        super().__init__(
            simulation_config=simulation_config,
            logger=logger,
            reset_cell_params_to_lower_bounds_on_init=reset_cell_params_to_lower_bounds_on_init,
        )

        self.model = None
        self.model_pool = None

    def optimize(self, target_V: torch.Tensor) -> torch.Tensor:
        # Get voltage with characteristics similar to target
        simulated_V_for_next_stage, param_samples_for_next_stage = self.match_voltage(
            target_V
        )

        n_slices = (
            self.config["optimization_parameters"]
            .get("parametric_distribution", {})
            .get("n_slices", 0)
        )
        simulations_per_amp = (
            self.config["optimization_parameters"]
            .get("parametric_distribution", {})
            .get("simulations_per_amp", 1)
        )
        # we should do a parametric sampling of parameters to train network
        if n_slices > 1:
            simulated_V_dist, param_samples_dist, simulated_amps = self.get_parametric_distribution(
                n_slices, simulations_per_amp
            )
            simulated_V_for_next_stage = torch.cat(
                (simulated_V_for_next_stage, simulated_V_dist), dim=1
            )
            param_samples_for_next_stage = torch.cat(
                (param_samples_for_next_stage, param_samples_dist), dim=1
            )
        else:
            print(f"Parametric distribution parameters 'n_slices' not set, skipping.")

        self.model = self.init_nn_model(
            in_channels=target_V.shape[1], out_channels=self.num_params
        )

        # Resample to match the length of target data
        resampled_data = self.resample_voltage(
            simulated_V_for_next_stage, target_V.shape[1]
        )

        lows = [p["low"] for p in self.config["optimization_parameters"]["params"]]
        highs = [p["high"] for p in self.config["optimization_parameters"]["params"]]

        # Train model
        self.train_model(resampled_data, param_samples_for_next_stage, lows, highs)

        # Predict and take max across ci to prevent underestimating
        predictions = self.predict_with_model(target_V, lows, highs)
        predictions = torch.max(predictions, dim=0).values

        return predictions

    def optimize_with_segregation(
        self, target_V: torch.Tensor, segregate_by: str = "voltage"
    ) -> torch.Tensor:
        if segregate_by == "voltage":
            cut_func = self.cut_voltage_region
        elif segregate_by == "time":
            cut_func = self.cut_time_region
        else:
            raise ValueError("segregate_by must be either 'voltage' or 'time'.")

        # Get voltage with characteristics similar to target
        simulated_V_for_next_stage, param_samples_for_next_stage = self.match_voltage(
            target_V
        )

        n_slices = (
            self.config["optimization_parameters"]
            .get("parametric_distribution", {})
            .get("n_slices", 0)
        )
        simulations_per_amp = (
            self.config["optimization_parameters"]
            .get("parametric_distribution", {})
            .get("simulations_per_amp", 1)
        )
        # we should do a parametric sampling of parameters to train network
        if n_slices > 1:
            simulated_V_dist, param_samples_dist, simulated_amps = self.get_parametric_distribution(
                n_slices, simulations_per_amp
            )
            simulated_V_for_next_stage = torch.cat(
                (simulated_V_for_next_stage, simulated_V_dist)
            )
            param_samples_for_next_stage = torch.cat(
                (param_samples_for_next_stage, param_samples_dist)
            )
        else:
            print(f"Parametric distribution parameters 'n_slices' not set, skipping.")

        # Resample to match the length of target data
        resampled_data = self.resample_voltage(
            simulated_V_for_next_stage, target_V.shape[1]
        )

        # Store original params to restore later
        orig_params = {
            p["channel"]: p for p in self.config["optimization_parameters"]["params"]
        }

        self.model_pool = []
        prediction_pool = []

        for segregation in self.config["segregation"]:
            seg_params = segregation["params"]
            bounds = segregation[segregate_by]

            # Set the parameters
            param_ind = [
                i for i, (c, _) in enumerate(orig_params.items()) if c in seg_params
            ]

            lows = [
                p["low"] for channel, p in orig_params.items() if channel in seg_params
            ]
            highs = [
                p["high"] for channel, p in orig_params.items() if channel in seg_params
            ]
            self.num_params = len(seg_params)

            # Cut the required segment
            cut_target_V = cut_func(resampled_data, bounds)

            # Train a model for this segment
            self.model = self.init_nn_model(
                in_channels=cut_target_V.shape[1], out_channels=self.num_params
            )
            self.train_model(
                cut_target_V,
                param_samples_for_next_stage[:, param_ind],
                lows=lows,
                highs=highs,
            )

            # Predict and take max across ci to prevent underestimating
            predictions = self.predict_with_model(cut_target_V, lows, highs)
            predictions = torch.max(predictions, dim=0).values

            # Update cell model
            self.cell.set_parameters(seg_params, predictions)

            # Update the pools
            prediction_pool.append(predictions)
            self.model_pool.append(self.model)

        prediction_pool = torch.cat(prediction_pool, dim=0)

        return prediction_pool

    def init_nn_model(self, in_channels: int, out_channels: int) -> torch.nn.Sequential:
        model = torch.nn.Sequential(
            torch.nn.Linear(in_channels, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, out_channels),
            torch.nn.Sigmoid(),
        )
        return model

    def train_model(
        self,
        voltage_data: torch.Tensor,
        target_params: torch.Tensor,
        lows,
        highs,
    ) -> None:
        optim = torch.optim.SGD(self.model.parameters(), lr=1e-8)
        sigmoid_mins = torch.tensor(lows)
        sigmoid_maxs = torch.tensor(highs)

        self.logger.info("Training a model with SGD optimizer and lr = 1e-8.")
        self.logger.info(
            f"Number of trainable parameters: {sum(p.numel() for p in self.model.parameters() if p.requires_grad)}"
        )

        self.model.train(True)
        for i in range(voltage_data.shape[0]):
            # Track loss to stop if it starts to go up
            loss0 = np.inf

            for ne in range(self.config["optimization_parameters"]["num_epochs"]):
                pred = (
                    self.model(voltage_data[i]) * (sigmoid_maxs - sigmoid_mins)
                    + sigmoid_mins
                )
                loss = torch.nn.functional.l1_loss(pred, target_params[i])

                if ne % 100 == 0:
                    self.logger.epoch(ne, "l1_loss", float(loss.detach().numpy()))

                if loss.detach().numpy() >= loss0:
                    break
                loss0 = loss.detach().numpy()

                loss.backward()
                optim.step()

    def predict_with_model(self, target_V: torch.Tensor, lows, highs) -> torch.Tensor:
        sigmoid_mins = torch.tensor(lows)
        sigmoid_maxs = torch.tensor(highs)

        self.model.eval()
        outs = []
        for i in range(target_V.shape[0]):
            out = self.model(target_V[i]) * (sigmoid_maxs - sigmoid_mins) + sigmoid_mins
            outs.append(out.reshape(1, -1))

        return torch.cat(outs, dim=0)

    def get_parametric_distribution(self, n_slices, simulations_per_amp) -> tuple:
        params = [
            p["channel"] for p in self.config["optimization_parameters"]["params"]
        ]
        lows = [p["low"] for p in self.config["optimization_parameters"]["params"]]
        highs = [p["high"] for p in self.config["optimization_parameters"]["params"]]
        steps = int(
            self.config["simulation_parameters"]["h_tstop"]
            / self.config["simulation_parameters"]["h_dt"]
        )

        param_samples_for_next_stage = torch.zeros(
            (self.num_ampl, simulations_per_amp, self.num_params)
        )
        simulated_V = torch.zeros(
            (
                self.num_ampl,
                simulations_per_amp,
                steps,
            )
        )

        param_dist = np.array(
            [
                np.arange(low, high, (high - low) / n_slices)
                for low, high in zip(lows, highs)
            ]
        ).T

        amps = self.config["optimization_parameters"]["amps"]
        print(
            f"Sampling parameter space... this may take a while. {len(amps)} amps * {simulations_per_amp} simulations per amp = {len(amps) * simulations_per_amp}"
        )
        s_amps = []
        for amp_ind, amp in enumerate(amps):
            print(
                f"    Generating {simulations_per_amp} simulations_per_amp at {amp:.2f} amps."
            )
            for slice_ind in range(simulations_per_amp):
                # For each current injection amplitude, sample random parameters
                param_inds = np.random.randint(
                    0, n_slices, len(params)
                )  # get random indices for our params
                param_sample = param_dist.T[
                    range(len(params)), param_inds
                ]  # select the params
                simulated_V[amp_ind][slice_ind] = self.simulate(
                    amp, params, param_sample
                )
                param_samples_for_next_stage[amp_ind][slice_ind] = torch.Tensor(
                    param_sample
                )

                s_amps.append(amp)

        s_v = torch.flatten(simulated_V).reshape(
            [len(amps) * simulations_per_amp, steps]
        )
        s_param = torch.flatten(param_samples_for_next_stage).reshape(
            [len(amps) * simulations_per_amp, len(params)]
        )
        s_amps = torch.tensor(s_amps)

        return s_v, s_param, s_amps

    def match_voltage(self, target_V: torch.Tensor) -> tuple:
        # Get target voltage summary features
        num_target_spikes, target_interspike_times = self.extract_summary_features(
            target_V
        )  # (N x 1)

        # Set a vector of non-correspondence to target features
        non_corresp = torch.ones(self.num_ampl)

        simulated_V_for_next_stage = torch.zeros(
            (
                self.num_ampl,
                int(
                    self.config["simulation_parameters"]["h_tstop"]
                    / self.config["simulation_parameters"]["h_dt"]
                ),
            )
        )
        param_samples_for_next_stage = torch.zeros((self.num_ampl, self.num_params))

        self.logger.info(
            f"Matching {self.config['optimization_parameters']['num_amps_to_match']} amplitudes."
        )

        params = [
            p["channel"] for p in self.config["optimization_parameters"]["params"]
        ]
        lows = [p["low"] for p in self.config["optimization_parameters"]["params"]]
        highs = [p["high"] for p in self.config["optimization_parameters"]["params"]]

        while torch.sum(non_corresp) > (
            self.num_ampl - self.config["optimization_parameters"]["num_amps_to_match"]
        ):
            # Simulate with these samples
            simulated_V = torch.zeros(
                (
                    self.num_ampl,
                    int(
                        self.config["simulation_parameters"]["h_tstop"]
                        / self.config["simulation_parameters"]["h_dt"]
                    ),
                )
            )
            for ind, amp in enumerate(self.config["optimization_parameters"]["amps"]):
                if non_corresp[ind] == 1:
                    # For each current injection amplitude, sample random parameters
                    param_samples = np.random.uniform(low=lows, high=highs)
                    simulated_V[ind] = self.simulate(amp, params, param_samples)
                    param_samples_for_next_stage[ind] = torch.Tensor(param_samples)

            # Compute summary features for simulated data
            (
                num_spikes_simulated,
                simulated_interspike_times,
            ) = self.extract_summary_features(simulated_V)

            # Save those voltage traces which match the target
            cond_spikes = self.get_match_condition(
                num_target_spikes,
                num_spikes_simulated,
                self.config["summary_features"]["mc_num_spikes"],
            )
            cond_times = self.get_match_condition(
                target_interspike_times,
                simulated_interspike_times,
                self.config["summary_features"]["mc_interspike_time"],
            )
            cond_min = self.get_match_condition(
                torch.min(target_V, dim=1).values,
                torch.min(simulated_V, dim=1).values,
                self.config["summary_features"]["mc_min_v"],
            )
            cond_mean = self.get_match_condition(
                torch.mean(target_V, dim=1),
                torch.mean(simulated_V, dim=1),
                self.config["summary_features"]["mc_mean_v"],
            )
            cond_max = self.get_match_condition(
                torch.max(target_V, dim=1).values,
                torch.max(simulated_V, dim=1).values,
                self.config["summary_features"]["mc_max_v"],
            )

            cond = cond_spikes & cond_times & cond_mean & cond_min & cond_max
            simulated_V_for_next_stage[cond] = simulated_V[cond]
            non_corresp[cond] = 0
            self.logger.info(
                f"Total amplitudes matched: {int(torch.sum(1 - non_corresp))}/{self.config['optimization_parameters']['num_amps_to_match']}."
            )

        simulated_V_for_next_stage = simulated_V_for_next_stage[non_corresp == 0]
        param_samples_for_next_stage = param_samples_for_next_stage[non_corresp == 0]
        inds_of_ampl_next_stage = np.where(non_corresp == 0)[0].tolist()

        self.logger.info(
            f"Matched amplitudes: {np.array(self.config['optimization_parameters']['amps'])[inds_of_ampl_next_stage]}"
        )

        return simulated_V_for_next_stage, param_samples_for_next_stage

    def get_match_condition(
        self, target: torch.Tensor, simulated: torch.Tensor, threshold: float
    ):
        return (torch.abs(target - simulated) < threshold).flatten()

    def extract_summary_features(self, V: torch.Tensor) -> tuple:
        threshold_crossings = torch.diff(
            V > self.config["summary_features"]["spike_threshold"], dim=1
        )
        num_spikes = torch.round(torch.sum(threshold_crossings, dim=1) * 0.5)

        interspike_times = torch.zeros((V.shape[0], 1))
        for i in range(threshold_crossings.shape[0]):
            interspike_times[i, :] = torch.mean(
                torch.diff(
                    torch.arange(threshold_crossings.shape[1])[
                        threshold_crossings[i, :]
                    ]
                ).float()
            )

        return num_spikes, interspike_times



from simulation_configs import pospischilsPYr, pospischilsPYr_passive, LA_A_seg

from act import simulator

if __name__ == "__main__":
    config = LA_A_seg

    p = simulator.run(config, subprocess=False)



import json
import os

import numpy as np
import torch

from act.act_types import SimulationConfig
from act.analysis import save_plot
from act.optim import ACTOptimizer


def get_voltage_trace_from_params(
    simulation_config: SimulationConfig,
) -> torch.Tensor:
    optim = ACTOptimizer(
        simulation_config=simulation_config, set_passive_properties=False
    )
    target_V = []
    output_folder = os.path.join(
        simulation_config["output"]["folder"], simulation_config["run_mode"], "target"
    )
    if not os.path.exists(output_folder):
        os.makedirs(output_folder, exist_ok=True)

    target_params = simulation_config["optimization_parameters"]["target_params"]
    dt = simulation_config["simulation_parameters"]["h_dt"]
    simulated_label = simulation_config["output"].get("simulated_label", "Simulated")
    target_label = simulation_config["output"].get("target_label", "Target")
    for i, amp in enumerate(simulation_config["optimization_parameters"]["amps"]):
        params = [
            p["channel"] for p in simulation_config["optimization_parameters"]["params"]
        ]
        tv = optim.simulate(amp, params, target_params).reshape(1, -1)
        target_V.append(tv)
        # write to output folder / mode / target
        if i % 5 == 0:  # 5 should be user defined
            save_plot(
                amp,
                output_folder,
                simulated_data=None,
                target_V=tv.detach().numpy(),
                output_file=f"target_{(amp * 1000):.0f}nA.png",
                dt=dt,
                simulated_label=simulated_label,
                target_label=target_label,
            )
    target_V = torch.cat(target_V, dim=0)

    # save passive properties
    passive_properties, passive_v = optim.calculate_passive_properties(
        params, target_params
    )
    with open(
        os.path.join(output_folder, "target_passive_properties.json"),
        "w",
        encoding="utf-8",
    ) as fp:
        json.dump(passive_properties, fp, indent=2)

    save_plot(
        -0.1,
        output_folder,
        simulated_data=passive_v,
        output_file="passive_-100nA.png",
        dt=dt,
        simulated_label=simulated_label,
    )

    return target_V



import os

import numpy as np
import pandas as pd
from simulation_configs import pospischilsPYr, pospischilsPYr_passive, LA_A_seg

from act import analysis


def main():
    # for mode in ["original", "segregated"]:
    #    pospischilsPYr["run_mode"] = mode
    #    analysis.print_run_stats(pospischilsPYr)

    analysis.print_run_stats(LA_A_seg)


if __name__ == "__main__":
    main()



"""
The purpose of this module is to assist in the creation of single cells, from scratch
"""

cell_hoc_template = """
{{load_file("stdrun.hoc")}}
{{load_file("nrngui.hoc")}}

begintemplate {cell_name}

	public soma
	create soma[1]

	public all, somatic
    objref all, somatic

	proc init() {{	
		all = new SectionList()
		somatic = new SectionList()
		soma[0] all.append()
		soma[0] somatic.append()

		{global_params}	
		
		/////// topology ////////
		// No connections to be made
		
		/////// geometry ////////
		soma[0] {{
            {geometry_section}
        }}

        //////// mechanisms /////////
		soma[0] {{	
			{mechanism_section}
		}}
				
	define_shape()// builtin fcn: fill in 3d info for sections defined
	}}//end init

endtemplate {cell_name}
"""


class CellGenerator:

    """
    Generate a cell

    eg:

    from act.builder import CellGenerator

    pyramidal_cell = CellGenerator('pyr_type_A')
    pyramidal_cell.set_global_params(Rm = 80000, Cm = 2.4, Ra = 150)
    pyramidal_cell.insert_mech('leak', el_leak = -72)
    pyramidal_cell.set_geometry(L = 25, diam = 25, nseg = 1)
    pyramidal_cell.write_hoc('cell_templates.hoc')

    """

    def __init__(self, cell_name):
        self.cell_name = cell_name
        self.global_params = {}
        self.mods = {}
        self.set_geometry()

    def set_geometry(self, L=25, diam=25, nseg=1):
        """
        Set geometric properties for the cell
        """
        self.L = L
        self.diam = diam
        self.nseg = nseg

    def set_global_params(self, *args, **kwargs):
        """
        Set global properties like Ra and Cm for the cell
        """
        for key, value in kwargs.items():
            self.global_params[key] = value

    def insert_mech(self, mechanism, **params):
        """
        Insert a new mechanism and properties, can update
        """
        self.mods[mechanism] = params

    def write_hoc(self, file_path, append=False):
        """
        Take the cell that has been built and write to a file_path

        params:
        file_path: where the files is written to
        append: if you have multiple templates that you're writting to the same file, set to True
        """

        mechanism_section = ""
        for mech, params in self.mods.items():
            mech_line = f"insert {mech} "
            for param, value in params.items():
                mech_line += f"{param} = {value} "
            mechanism_section = mechanism_section + mech_line + "\n"

        global_params = ""
        for param, value in self.global_params.items():
            global_params += f"{param} = {value}\n"

        geometry_section = f"""
        L = {self.L}
        diam = {self.diam}
        nseg = {self.nseg}
        """

        cell_hoc = cell_hoc_template.format(
            cell_name=self.cell_name,
            global_params=global_params,
            geometry_section=geometry_section,
            mechanism_section=mechanism_section,
        )

        # write to file
        with open(file_path, "a" if append else "w") as text_file:
            text_file.write(cell_hoc)



import json
import os
import shutil
from multiprocessing import Process

import h5py
import numpy as np
import pandas as pd
import torch
from neuron import h

from act.act_types import SimulationConfig
from act.analysis import save_mse_corr, save_plot, save_prediction_plots
from act.logger import ACTLogger
from act.metrics import correlation_score, mse_score
from act.optim import GeneralACTOptimizer
from act.target_utils import get_voltage_trace_from_params


def _run(config: SimulationConfig):
    if config["optimization_parameters"]["num_epochs"] < 1000:
        raise ValueError("Number of epochs is expected to be >= 1000.")

    output_folder = config["output"]["folder"]
    run_output_folder_name = f"{config['run_mode']}"

    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    output_folder = os.path.join(config["output"]["folder"], run_output_folder_name)
    # delete old results
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder, ignore_errors=True)
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    os.system(f"nrnivmodl {config['cell']['modfiles_folder']}")

    logger = ACTLogger()
    logger.info(
        f"Number of amplitudes: {len(config['optimization_parameters']['amps'])}"
    )

    try:
        h.nrn_load_dll("./x86_64/.libs/libnrnmech.so")
    except:
        logger.info("Mod files already loaded. Continuing.")

    # Get target voltage
    if config["optimization_parameters"]["target_V"] is not None:
        target_V = config["optimization_parameters"]["target_V"]
    elif config["optimization_parameters"]["target_params"] is not None:
        target_V = get_voltage_trace_from_params(config)
    else:
        raise ValueError(
            "Must specify either target_V or target_params for optimization_parameters"
        )

    logger.info(f"Target voltage shape: {target_V.shape}")

    # Run the optimizer
    pred_pool = []
    err_pool = []
    params = [p["channel"] for p in config["optimization_parameters"]["params"]]
    for _ in range(config["optimization_parameters"]["num_repeats"]):
        if config["run_mode"] == "original":
            optim = GeneralACTOptimizer(simulation_config=config, logger=logger)
            predictions = optim.optimize(target_V)
        elif config["run_mode"] == "segregated":
            optim = GeneralACTOptimizer(simulation_config=config, logger=logger)
            predictions = optim.optimize_with_segregation(target_V, "voltage")
        else:
            raise ValueError(
                "run mode not specified, 'original' or 'segregated' supported."
            )

        pred_pool.append(predictions)

        sims = []
        for amp in config["optimization_parameters"]["amps"]:
            sims.append(optim.simulate(amp, params, predictions).reshape(1, -1))
        sims = torch.cat(sims, dim=0)

        # Compute composite error
        error = mse_score(target_V, sims) + (1 - abs(correlation_score(target_V, sims)))
        err_pool.append(error)

    predictions = pred_pool[np.argmin(err_pool)]

    with open(os.path.join(output_folder, "config.json"), "w") as file:
        json.dump(config, file, indent=2)

    # Save predictions
    pred_df = pd.DataFrame(dict(zip(params, predictions.detach().numpy())), index=[0])

    g_leak_var = optim.cell.gleak_var
    g_bar_leak = optim.cell.g_bar_leak
    if g_leak_var and g_bar_leak:  # if the user set the passive properties
        pred_df.insert(0, g_leak_var, [g_bar_leak])

    pred_df.to_csv(os.path.join(output_folder, "pred.csv"))
    save_mse_corr(target_V, config, predictions, output_folder)

    # save passive properties
    passive_properties, passive_v = optim.calculate_passive_properties(
        params, predictions.detach().numpy()
    )
    with open(
        os.path.join(output_folder, "pred_passive_properties.json"),
        "w",
        encoding="utf-8",
    ) as fp:
        json.dump(passive_properties, fp, indent=2)

    if config["output"]["produce_plots"]:
        # output passive amps don't matter
        dt = config["simulation_parameters"]["h_dt"]
        simulated_label = config["output"].get("simulated_label", "Simulated")
        save_plot(
            -0.1,
            output_folder,
            simulated_data=passive_v,
            output_file="passive_-100nA.png",
            dt=dt,
            simulated_label=simulated_label,
        )

        i = 0
        amp_out = []
        simulated_V_out = []
        target_V_out = []
        while i < len(config["optimization_parameters"]["amps"]):
            amp_i = config["optimization_parameters"]["amps"][i]
            target_Vi = target_V[i].reshape(1, -1)
            simulated_Vi = save_prediction_plots(
                target_Vi,
                config["optimization_parameters"]["amps"][i],
                config,
                predictions,
                output_folder,
            )
            simulated_V_out.append(simulated_Vi)
            target_V_out.append(target_Vi)
            amp_out.append(amp_i)

            i += 5  # should be user variable

        # Save the voltage traces for debugging
        target_V_list = [list(t.detach().numpy()[0]) for t in target_V_out]
        simulated_V_list = [list(t.detach().numpy()[0]) for t in simulated_V_out]

        f = h5py.File(os.path.join(output_folder, "traces.h5"), "w")
        target_grp = f.create_group("target")
        simulated_grp = f.create_group("simulated")

        target_grp.create_dataset(
            "voltage_trace",
            (len(target_V_list), len(target_V_list[0])),
            dtype="f",
            data=target_V_list,
        )
        simulated_grp.create_dataset(
            "voltage_trace",
            (len(simulated_V_list), len(simulated_V_list[0])),
            dtype="f",
            data=simulated_V_list,
        )
        f.create_dataset("amps", (len(amp_out)), dtype="f", data=amp_out)
        f.close()


def run(config: SimulationConfig, subprocess=True):
    try:
        if subprocess:
            p = Process(target=_run, args=[config])
            p.start()
            p.join()
            p.terminate()
        else:
            _run(config)
    except:
        raise
    finally:  # always remove this folder
        if os.path.exists("x86_64"):
            os.system("rm -r x86_64")



import torch

# We can potentially add CNN layers before Linear layers


# @CHECK
class BranchingNet(torch.nn.Module):
    def __init__(self, in_channels, out_channels, summary_features):
        super().__init__()
        self.summary_features = summary_features
        self.voltage_branch = torch.nn.Sequential(
            torch.nn.Linear(in_channels, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, out_channels),
        )
        self.summary_branch = torch.nn.Sequential(
            torch.nn.Linear(summary_features.shape[-1], out_channels)
        )
        self.sigmoid = torch.nn.Sigmoid()

    def forward(self, X):
        voltage_out = self.voltage_branch(X)
        summary_out = self.summary_branch(self.summary_features)
        return self.sigmoid(voltage_out + summary_out)


# @CHECK
class EmbeddingNet(torch.nn.Module):
    def __init__(self, in_channels, out_channels, summary_features):
        super().__init__()
        self.summary_features = summary_features
        self.embedder = torch.nn.Sequential(
            torch.nn.Linear(in_channels, 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 64),
            torch.nn.ReLU(),
        )
        self.predictor = torch.nn.Sequential(
            torch.nn.Linear(64 + summary_features.shape[-1], 256),
            torch.nn.ReLU(),
            torch.nn.Linear(256, 256),
            torch.nn.Sigmoid(),
        )

    def forward(self, X):
        embedding = self.embedder(X)
        return self.predictor(torch.cat(embedding, self.summary_features))



from datetime import datetime


class ACTDummyLogger:
    def __init__(self):
        pass

    def info(self, msg: str):
        pass

    def epoch(self, epoch_num: int, loss_name: str, loss_value: float):
        pass


class ACTLogger:
    def __init__(self):
        print(f"\n({datetime.now()})-[START]")

    def info(self, msg: str) -> None:
        print(f"({datetime.now()})-[INFO]: {msg}")

    def epoch(self, epoch_num: int, loss_name: str, loss_value: float) -> None:
        print(
            f"({datetime.now()})-[EPOCH {epoch_num}]: {loss_name} = {round(loss_value, 5)}"
        )



import torch


def correlation_score(target_data: torch.Tensor, simulated_data: torch.Tensor) -> float:
    cov = (target_data - torch.mean(target_data, dim=1, keepdim=True)) * (
        simulated_data - torch.mean(simulated_data, dim=1, keepdim=True)
    )
    cov = torch.sum(cov, dim=1)

    var0 = torch.sum(
        torch.square(target_data - torch.mean(target_data, dim=1, keepdim=True)), dim=1
    )
    var1 = torch.sum(
        torch.square(simulated_data - torch.mean(simulated_data, dim=1, keepdim=True)),
        dim=1,
    )
    corr = cov / (torch.sqrt(var0 * var1) + 1e-15)

    return float(torch.mean(corr).detach())


def mse_score(target_data: torch.Tensor, simulated_data: torch.Tensor) -> float:
    return float(
        torch.mean(
            torch.mean(torch.square(target_data - simulated_data), dim=1)
        ).detach()
    )



from neuron import h

from act import act_types


class CellModel:
    def __init__(self, hoc_file: str, cell_name: str):
        """
        Constructs a Cell object.

        Parameters:
        ----------
        hoc_file: str
            Name of the .hoc file that defines the cell.

        cell_name: str
            Name of the cell in the .hoc file.
        """

        # Load the .hoc file
        h.load_file(hoc_file)

        # Initialize the cell
        hoc_cell = getattr(h, cell_name)()
        self.all = hoc_cell.all
        self.soma = hoc_cell.soma

        # Initialize recorders
        self.t = h.Vector().record(h._ref_t)
        self.Vm = h.Vector().record(self.soma[0](0.5)._ref_v)

        # Init injection
        self.inj = h.IClamp(self.soma[0](0.5))
        self.I = h.Vector().record(h._ref_i)

        # Passive properties
        self.gleak_var = None
        self.g_bar_leak = None

    def set_parameters(self, parameter_list: list, parameter_values: list) -> None:
        for sec in self.all:
            for index, key in enumerate(parameter_list):
                setattr(sec, key, parameter_values[index])

    def apply_current_injection(self, amp: float, dur: float, delay: float) -> None:
        self.inj.amp = amp
        self.inj.dur = dur
        self.inj.delay = delay

    def set_passive_properties(
        self, passive_properties: act_types.PassiveProperties
    ) -> None:
        if passive_properties:
            v_rest = passive_properties.get("v_rest")  # mV
            r_in = passive_properties.get("r_in")  # Ohms
            tau = passive_properties.get("tau")  # ms

            gleak_var = passive_properties.get("leak_conductance_variable")
            eleak_var = passive_properties.get("leak_reversal_variable")

            # Assuming Vrest is within the range for ELeak
            # ELeak = Vrest
            if v_rest and eleak_var:
                print(f"Setting {eleak_var} = {v_rest}")
                self.set_parameters([eleak_var], [v_rest])
            else:
                print(
                    f"Skipping analytical setting of e_leak variable. Cell v_rest and/or leak_reversal_variable not specified in config."
                )

            for sec in self.all:
                # Rin = 1/(Area*g_bar_leak)
                # g_bar leak = 1/(Rin*Area)
                area = sec(
                    0.5
                ).area()  # have to specify the location to get access to area
                if r_in:
                    g_bar_leak = 1 / (r_in * area) * 1e2  # area m to cm?
                    print(f"Setting {sec}.{gleak_var} = {g_bar_leak:.8f}")
                    setattr(sec, gleak_var, g_bar_leak)
                    if sec == self.soma:
                        g_bar_leak = g_bar_leak

                    if tau:
                        # tau = (R*C) = R*Area*cm
                        # cm = tau*g_bar_leak*Area
                        # cm/cm2 = (tau*g_bar_leak*Area)/Area = tau*g_bar_leak
                        cm = tau * g_bar_leak * 1e3  # tau ms->s
                        print(f"Setting {sec}.cm = {cm:.8f}")
                        setattr(sec, "cm", cm)
                    else:
                        print(
                            f"Skipping analytical setting of cm variable. Cell tau not specified in config."
                        )
                else:
                    print(
                        f"Skipping analytical setting of gleak and cm variables. Cell r_in not specified in config."
                    )
        else:
            print(
                f"Skipping analytical setting of passive properties, no cell passive_properties specified in config."
            )
